//
//  coval.c
//
//  coval version 0.8
//      2012.11.01 Takeshi Ogihara
//  Copyright (C) 2010, 2011, 2012 Takeshi Ogihara
//
#include <limits.h>
#include "coval.h"

#if USE_COVAL_BRIDGE
struct _bridge_base;
/* Local */ void call_coval_bridge(coval_base *cov, struct _bridge_base *brg);
#define Coval_bridge_ID  (coval_callback)(1)    // special value
#define _makeDiffData(A)  *(char *)&(A)->_d = *(char *)(A)->_p + 1
#endif

#define isCenter(Cp)    ((Cp)->_p == &((Cp)->_d))

void _bindCoval(coval_base *target, coval_base *cov, void *pp)
{
    if (target == NULL || cov == NULL || target == cov || cov->_p != NULL)
        return;

    coval_base *center;
    if (target->_p == NULL) { // first bind
        center = target;
        center->_p = &(center->_d); // make center coval
        center->_b._cinfo._count = 1;
        center->_b._cinfo._properties = 0; // default values
        center->_b._cinfo._inLoop = 0;
        center->_nx = center;  // simple ring
        center->_loopcntl = center;
        center->_upptr = center->_up ? center : NULL;
    }else {
        center = isCenter(target) ? target : target->_b._cent;
        center->_b._cinfo._count++;
    }
    cov->_p = (pp != NULL) ? pp : center->_p; // pp points a certain member
    cov->_b._cent = center;
    if (cov->_up)
        center->_upptr = cov;

    /* make a chain */
    coval_base *w = center->_loopcntl;  // previous last element
    cov->_nx = center;
    center->_loopcntl = w->_nx = cov;
#if USE_COVAL_BRIDGE
    _makeDiffData(cov);
    // Bridge: make different value in the bridge
#endif
}

void _unbindCoval(coval_base *cov, unsigned int bsize)
{
    coval_base *center;
    unsigned char *tgt, *src;
    unsigned int i, memsize;

    if (cov == NULL || cov->_p == NULL)
        return; // false
    center = isCenter(cov) ? cov : cov->_b._cent;
    if (center->_b._cinfo._inLoop)
        return; // false
    memsize = bsize - (sizeof(coval_base) - sizeof(long));
    src = (unsigned char *)(center->_p); // in covalbridge, _p != &_d, maybe

    if (center->_b._cinfo._count == 1) { // only 2 are in the ring
        coval_base *del = center->_nx;
        tgt = (unsigned char *)&(del->_d);
        center->_p = del->_p = NULL;
        center->_b._cent = del->_b._cent = NULL;
    }else {
        coval_base *p, *upwk;
        upwk = center->_upptr;
        if (cov == center) {
            coval_base *tail = center->_loopcntl; // last element
            coval_base *nw = center->_nx;
            tgt = (unsigned char *)&(nw->_d);
            tail->_nx = nw;
            nw->_loopcntl = tail;
            nw->_p = &(nw->_d);
            nw->_b._cinfo = center->_b._cinfo;
            nw->_b._cinfo._count--;
            for (p = nw->_nx; p != nw; p = p->_nx) {
                p->_b._cent = nw;
                p->_p = (void *)tgt;
            }
            nw->_upptr = center->_upptr;
            center = nw;
        }else {
            coval_base *q = center; // previous one
            for (p = center->_nx; p != center; q = p, p = p->_nx) {
                if (p == cov) {
                    q->_nx = p->_nx;
                    center->_b._cinfo._count--;
                    if (p->_nx == center)
                        center->_loopcntl = q;
                    break;
                }
            }
            tgt = (unsigned char *)&(cov->_d);
        }
        cov->_p = NULL;
        cov->_b._cent = NULL;
        if (upwk == cov) { // search new update function
            upwk = center->_up ? center : NULL;
            for (p = center->_nx; p != center; p = p->_nx) {
                if (p->_up)
                    upwk = p;
            }
            center->_upptr = upwk;
        }
    }

    for (i = 0; i < memsize; i++)
        *tgt++ = *src++;
}

void _coval_call(coval_base *caller, int metoo)
{
    if (caller->_p == NULL)
        return;

    int ret, breakable;
    coval_base *center;

    center = isCenter(caller) ? caller : caller->_b._cent;
    if (center->_b._cinfo._inLoop == 0) {
        center->_nx->_loopcntl = caller->_nx;
        center->_b._cinfo._inLoop = 1;
    }else if (caller->_ringflag == 1)
        return; // nest > 1
    else
        center->_b._cinfo._inLoop++;
    caller->_ringflag = 1;

    breakable = (center->_b._cinfo._properties & BIND_FLAG_BREAKABLE) != 0;

    do {
        coval_base *loopptr = center->_nx->_loopcntl;
        // this calling loop may proceed during propagation
        center->_nx->_loopcntl = loopptr->_nx;
        loopptr->_ringflag = 0;
        if (loopptr->_cb) {
            if (loopptr != caller || metoo) {
#if USE_COVAL_BRIDGE
                if (loopptr->_cb == Coval_bridge_ID) { /* Bridge */
                    call_coval_bridge(loopptr, loopptr->_ag);
                }else
#endif
                {
                    ret = loopptr->_cb(loopptr, loopptr->_ag);
                    if (breakable && ret == COVAL_PROC_SUCCESS)
                        break;
                }
            }
        }
    }while (caller->_ringflag == 1);
    center->_b._cinfo._inLoop--;
}

int _coval_request(coval_base *caller)
{
    if (caller && caller->_p) { // connected
        coval_base *center, *fp;
        center = isCenter(caller) ? caller : caller->_b._cent;
        if ((fp = center->_upptr) != NULL)
            fp->_up(fp, fp->_au, (void *)(caller->_p));
        return 1;
    }
    if (caller->_up)
        caller->_up(caller, caller->_au, (void *)&(caller->_d));
    return 0;
}

void _setBindingProperty(coval_base *cov, unsigned int prop, int flag)
{
    unsigned int b;
    if (cov->_p == NULL)
        return; /* not bound */
    coval_base *center = isCenter(cov) ? cov : cov->_b._cent;
    b = center->_b._cinfo._properties;
    if (flag) b |= prop;
    else b &= 0xff ^ prop;
    center->_b._cinfo._properties = b;
}

